/**********************************************************************
Copyright (c) 2008 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
2008 Andy Jefferson - restructured to take in applyXXX flags and evaluate different parts if required
    ...
**********************************************************************/
package org.datanucleus.query.evaluator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.ObjectManagerFactoryImpl;
import org.datanucleus.query.QueryUtils;
import org.datanucleus.query.compiler.QueryCompilation;
import org.datanucleus.query.evaluator.memory.InMemoryExpressionEvaluator;
import org.datanucleus.query.expression.CreatorExpression;
import org.datanucleus.query.expression.Expression;
import org.datanucleus.query.expression.InvokeExpression;
import org.datanucleus.query.expression.OrderExpression;
import org.datanucleus.query.symbol.PropertySymbol;
import org.datanucleus.query.symbol.Symbol;
import org.datanucleus.query.symbol.SymbolTable;
import org.datanucleus.store.query.Query;
import org.datanucleus.util.Localiser;
import org.datanucleus.util.NucleusLogger;
import org.datanucleus.util.StringUtils;

/**
 * Class to evaluate a Java "string-based" query in whole or part.
 * Takes in a list of instances and applies the required restrictions to the input giving a final result.
 * Typically extended for JDOQL, JPQL, SQL specifics.
 */
public abstract class JavaQueryEvaluator
{
    /** Localisation utility for output messages */
    protected static final Localiser LOCALISER = Localiser.getInstance(
        "org.datanucleus.Localisation", ObjectManagerFactoryImpl.class.getClassLoader());

    protected final String language;

    protected String candidateAlias = "this";

    protected Expression filter;

    protected Expression[] result;

    protected Expression[] ordering;

    protected Expression[] grouping;

    protected Expression having;

    protected SymbolTable symtbl;

    protected InMemoryExpressionEvaluator evaluator;

    /** List of candidates objects. */
    protected List candidates;

    /** Underlying "string-based" query. */
    protected Query query;

    /** Compilation of the underlying query. */
    protected QueryCompilation compilation;

    protected ClassLoaderResolver clr;

    /**
     * Constructor for the evaluator of a query in the specified language.
     * @param language Name of the language
     * @param query The underlying query
     * @param compilation Query compilation
     * @param clr ClassLoader resolver
     * @param candidates Candidate objects
     */
    public JavaQueryEvaluator(String language, Query query, QueryCompilation compilation, 
            ClassLoaderResolver clr, List candidates)
    {
        this.language = language;
        this.query = query;
        this.compilation = compilation;
        this.clr = clr;
        this.candidates = candidates;

        this.filter = compilation.getExprFilter();
        this.having = compilation.getExprHaving();
        this.ordering = compilation.getExprOrdering();
        this.grouping = compilation.getExprGrouping();
        this.result = compilation.getExprResult();
        this.symtbl = compilation.getSymbolTable();
        this.candidateAlias = (compilation.getCandidateAlias() != null ? 
                compilation.getCandidateAlias() : this.candidateAlias);

        evaluator = new InMemoryExpressionEvaluator(query.getObjectManager().getOMFContext().getQueryManager(),
            symtbl, query.getParsedImports(), clr, candidateAlias);
    }

    /**
     * Method to evaluate a subquery of the query being evaluated.
     * @param subquery The subquery
     * @param candidates The candidates for the subquery
     * @param compilation The subquery compilation
     * @return The result
     */
    protected abstract Collection evaluateSubquery(Query subquery, List candidates, QueryCompilation compilation);

    /**
     * Method to perform the evaluation, applying the query restrictions that are required.
     * @param applyFilter Whether to apply any filter constraints on the results
     * @param applyOrdering Whether to apply any order constraints on the results
     * @param applyResult Whether to apply any result/grouping/having on the results
     * @param applyResultClass Whether to apply any resultClass constraint on the results
     * @param applyRange Whether to apply any range constraint on the results
     * @return The results after evaluation.
     */
    public Collection execute(boolean applyFilter, boolean applyOrdering, boolean applyResult, 
            boolean applyResultClass, boolean applyRange)
    {
        if (candidates.isEmpty())
        {
            // No candidates so all match the query!
            return candidates;
        }
        else if (!applyFilter && !applyOrdering && !applyResult && !applyResultClass && !applyRange)
        {
            // Nothing to evaluate in-memory
            return candidates;
        }

        String[] subqueryAliases = compilation.getSubqueryAliases();
        if (subqueryAliases != null)
        {
            for (int i=0;i<subqueryAliases.length;i++)
            {
                // Evaluate subquery first
                Query subquery = query.getSubqueryForVariable(subqueryAliases[i]).getQuery();
                QueryCompilation subqueryCompilation = 
                    compilation.getCompilationForSubquery(subqueryAliases[i]);
                Collection subqueryResult = evaluateSubquery(subquery, candidates, subqueryCompilation);
                NucleusLogger.QUERY.debug(">> Subquery " + subqueryAliases[i] + " evaluates to " + 
                    StringUtils.collectionToString(subqueryResult));

                Symbol subquerySym = symtbl.getSymbol(subqueryAliases[i]);
                if (QueryUtils.queryReturnsSingleRow(subquery))
                {
                    // Subquery is expected to return single value
                    subquerySym.setValue(subqueryResult.iterator().next());
                }
                else
                {
                    subquerySym.setValue(subqueryResult);
                }
            }
        }

        List resultSet = new ArrayList(candidates);
        if (applyFilter && filter != null)
        {
            // Process any filter constraints
            if (NucleusLogger.QUERY.isDebugEnabled())
            {
                NucleusLogger.QUERY.debug(LOCALISER.msg("021012", "filter", language, filter));
            }
            resultSet = handleFilter(resultSet);
        }

        if (applyOrdering && ordering != null)
        {
            // Process any ordering constraints
            if (NucleusLogger.QUERY.isDebugEnabled())
            {
                NucleusLogger.QUERY.debug(LOCALISER.msg("021012", "ordering", language, 
                    StringUtils.objectArrayToString(ordering)));
            }
            resultSet = ordering(resultSet);
        }

        if (applyRange && 
            (query.getRangeFromIncl() != 0 || query.getRangeToExcl() != Long.MAX_VALUE || query.getRange() != null))
        {
            // Process any range constraints
            long fromIncl = query.getRangeFromIncl();
            long toExcl = query.getRangeToExcl();
            if (query.getRange() != null)
            {
                // Extract fromIncl/toExcl from range string
                int sepPos = query.getRange().indexOf(',');
                if (sepPos > 0)
                {
                    fromIncl = new Long(query.getRange().substring(0, sepPos).trim()).longValue();
                    toExcl = new Long(query.getRange().substring(sepPos+1).trim()).longValue();
                }
                else
                {
                    fromIncl = new Long(query.getRange().trim()).longValue();
                }
            }
            if (NucleusLogger.QUERY.isDebugEnabled())
            {
                NucleusLogger.QUERY.debug(LOCALISER.msg("021012", "range", language, "" + fromIncl + "," + toExcl));
            }
            resultSet = handleRange(resultSet, fromIncl, toExcl);
        }

        if (applyResult && result != null)
        {
            // Process any result/grouping/having constraints
            if (NucleusLogger.QUERY.isDebugEnabled())
            {
                NucleusLogger.QUERY.debug(LOCALISER.msg("021012", "result", language,
                    StringUtils.objectArrayToString(result)));
            }
            List aggregateList = new ArrayList();
            List s = resultSet;
            if (grouping != null)
            {
                s = sortByGrouping(resultSet);
            }

            aggregateList = s;
            if (grouping != null)
            {
                aggregateList = handleAggregates(s);
            }

            resultSet = handleResult(aggregateList);

            if (query.getResultDistinct())
            {
                List tmpList = new ArrayList();
                Iterator iter = resultSet.iterator();
                while (iter.hasNext())
                {
                    Object obj = iter.next();
                    if (!tmpList.contains(obj)) // Omit dups
                    {
                        tmpList.add(obj);
                    }
                }
                resultSet = tmpList;
            }
        }

        if (applyResultClass && query.getResultClass() != null)
        {
            // Process any result class constraints
            if (NucleusLogger.QUERY.isDebugEnabled())
            {
                NucleusLogger.QUERY.debug(LOCALISER.msg("021012", "resultClass", language, query.getResultClass().getName()));
            }
            if (!(result[0] instanceof CreatorExpression))
            {
                return this.mapResultClass(resultSet);
            }
        }

        return resultSet;
    }

    private List handleFilter(List set)
    {
        if (filter == null)
        {
            return set;
        }
    
        List result = new ArrayList();
        Iterator it = set.iterator();
        while (it.hasNext())
        {
            // Set the value of the candidate being tested, and evaluate it
            Object obj = it.next();
            symtbl.getSymbol(candidateAlias).setValue(obj);
            if (filter.evaluate(
                new InMemoryExpressionEvaluator(query.getObjectManager().getOMFContext().getQueryManager(), 
                    symtbl, query.getParsedImports(), clr, candidateAlias)) == Boolean.TRUE)
            {
                result.add(obj);
            }
        }
        return result;
    }

    private List handleRange(List set, long fromIncl, long toExcl)
    {
        if (toExcl - fromIncl <= 0)
        {
            return Collections.EMPTY_LIST;
        }
        else
        {
            List resultList = new ArrayList();
            Iterator it = set.iterator();
            // skipping the unnecessary objects
            for (long l = 0; l < fromIncl && it.hasNext(); l++)
            {
                it.next();
            }
            long l = 0;
            while (l < (toExcl - fromIncl) && it.hasNext())
            {
                resultList.add(it.next());
                l++;
            }
            return resultList;
        }
    }

    private List sortByGrouping(List set)
    {
        Object[] o = set.toArray();
        Arrays.sort(o, new Comparator()
        {
            public int compare(Object arg0, Object arg1)
            {
                for (int i=0; i<grouping.length; i++)
                {
                    symtbl.getSymbol(candidateAlias).setValue(arg0);
                    Object a = grouping[i].evaluate(evaluator);
                    symtbl.getSymbol(candidateAlias).setValue(arg1);
                    Object b = grouping[i].evaluate(evaluator);
                    int result = ((Comparable)a).compareTo(b);
                    if (result != 0)
                    {
                        return result;
                    }
                }
                return 0;
            }
        });
        return Arrays.asList(o);
    }

    private List ordering(List set)
    {
        if (ordering == null)
        {
            return set;
        }

        Symbol symbol = new PropertySymbol(Symbol.RESULTS_SET);
        if (symtbl.hasSymbol(Symbol.RESULTS_SET))
        {
            symbol = symtbl.getSymbol(Symbol.RESULTS_SET);
        }
        else
        {
            symtbl.addSymbol(symbol);
        }
        symbol.setValue(set);

        Object[] o = set.toArray();
        Arrays.sort(o, new Comparator()
        {
            public int compare(Object arg0, Object arg1)
            {
                for (int i=0; i<ordering.length; i++)
                {
                    symtbl.getSymbol(candidateAlias).setValue(arg0);
                    Object a = ordering[i].evaluate(
                        new InMemoryExpressionEvaluator(query.getObjectManager().getOMFContext().getQueryManager(),
                            symtbl, query.getParsedImports(), clr, candidateAlias));

                    symtbl.getSymbol(candidateAlias).setValue(arg1);
                    Object b = ordering[i].evaluate(
                        new InMemoryExpressionEvaluator(query.getObjectManager().getOMFContext().getQueryManager(),
                            symtbl, query.getParsedImports(), clr, candidateAlias));

                    int result = ((Comparable)a).compareTo(b);
                    if (result != 0)
                    {
                        OrderExpression orderExpr = (OrderExpression)ordering[i];
                        if (orderExpr.getSortOrder() == null || orderExpr.getSortOrder().equals("ascending"))
                        {
                            // Ascending
                            return result;
                        }
                        else
                        {
                            // Descending
                            return -1 * result;
                        }
                    }
                }
                return 0;
            }
        });
        return Arrays.asList(o);
    }

    /**
     * Checks if there are aggregates and handle it.
     * @param resultSet The resultSet containing all elements
     * @return A list with aggregated elements
     */
    private List handleAggregates(List resultSet)
    {
        Comparator c = new Comparator()
        {
            public int compare(Object arg0, Object arg1)
            {
                for (int i=0; i<grouping.length; i++)
                {
                    symtbl.getSymbol(candidateAlias).setValue(arg0);
                    Object a = grouping[i].evaluate(evaluator);
                    symtbl.getSymbol(candidateAlias).setValue(arg1);
                    Object b = grouping[i].evaluate(evaluator);
                    int result = ((Comparable)a).compareTo(b);
                    if (result != 0)
                    {
                        return result;
                    }
                }
                return 0;
            }
        };
    
        List groups = new ArrayList();
        List group = new ArrayList();
        groups.add(group);
        for (int i=0; i<resultSet.size(); i++)
        {
            if (i > 0)
            {
                if (c.compare(resultSet.get(i-1),resultSet.get(i)) != 0)
                {
                    group = new ArrayList();
                    groups.add(group);
                }
            }
            group.add(resultSet.get(i));
        }
        List result = new ArrayList();
        if (having != null)
        {
            for(int i=0; i<groups.size(); i++)
            {
                if (satisfiesHavingClause((List) groups.get(i)))
                {
                    result.addAll((Collection) groups.get(i));
                }
            }
        }
        else
        {
            for (int i = 0; i < groups.size(); i++)
            {
                result.addAll((Collection) groups.get(i));
            }
        }
        return result;
    }

    /**
     * Checks if the results set fulfils the having clause.
     * @param set Set of results
     * @return true if fulfilling having clause
     */
    private boolean satisfiesHavingClause(List set)
    {
        Symbol symbol = new PropertySymbol(Symbol.RESULTS_SET);
        if (symtbl.hasSymbol(Symbol.RESULTS_SET))
        {
            symbol = symtbl.getSymbol(Symbol.RESULTS_SET);
        }
        else
        {
            symtbl.addSymbol(symbol);
        }
        symbol.setValue(set);
        if (having.evaluate(evaluator) == Boolean.TRUE)
        {
            return true;
        }
        return false;
    }

    /**
     * Checks if there are aggregates and handles it.
     * @param resultSet The resultSet containing all elements
     * @return A list with aggregated elements
     */
    private List handleResult(List resultSet)
    {
        List result = new ArrayList();
        if (grouping != null)
        {
            Comparator c = new Comparator()
            {
                public int compare(Object arg0, Object arg1)
                {
                    if (grouping != null)
                    {
                        for (int i = 0; i < grouping.length; i++)
                        {
                            symtbl.getSymbol(candidateAlias).setValue(arg0);
                            Object a = grouping[i].evaluate(evaluator);
                            symtbl.getSymbol(candidateAlias).setValue(arg1);
                            Object b = grouping[i].evaluate(evaluator);
                            int result = ((Comparable) a).compareTo(b);
                            if (result != 0)
                            {
                                return result;
                            }
                        }
                    }
                    return 0;
                }
            };
    
            List groups = new ArrayList();
            List group = new ArrayList();
            if (resultSet.size() > 0)
            {
                groups.add(group);
            }
            for (int i = 0; i < resultSet.size(); i++)
            {
                if (i > 0)
                {
                    if (c.compare(resultSet.get(i - 1), resultSet.get(i)) != 0)
                    {
                        group = new ArrayList();
                        groups.add(group);
                    }
                }
                group.add(resultSet.get(i));
            }

            // Apply the result to the generated groups
            for (int i = 0; i < groups.size(); i++)
            {
                group = (List)groups.get(i);
                result.add(result(group));
            }
        }
        else
        {
            boolean aggregates = false;
            if (this.result.length > 0 && this.result[0] instanceof CreatorExpression)
            {
                Expression[] resExpr = (Expression[]) ((CreatorExpression) this.result[0]).getParameters().toArray(
                    new Expression[((CreatorExpression) this.result[0]).getParameters().size()]);
                for (int i = 0; i < resExpr.length; i++)
                {
                    if (resExpr[i] instanceof InvokeExpression)
                    {
                        String method = ((InvokeExpression) resExpr[i]).getOperation().toLowerCase();
                        if (method.equals("count") || method.equals("sum") || method.equals("avg") || 
                            method.equals("min") || method.equals("max"))
                        {
                            aggregates = true;
                        }
                    }
                }
            }
            else
            {
                for (int i = 0; i < this.result.length; i++)
                {
                    if (this.result[i] instanceof InvokeExpression)
                    {
                        String method = ((InvokeExpression) this.result[i]).getOperation().toLowerCase();
                        if (method.equals("count") || method.equals("sum") || method.equals("avg") || 
                            method.equals("min") || method.equals("max"))
                        {
                            aggregates = true;
                        }
                    }
                }
            }
    
            if (aggregates)
            {
                result.add(result(resultSet));
            }
            else
            {
                for (int i = 0; i < resultSet.size(); i++)
                {
                    result.add(result(resultSet.get(i)));
                }
            }
        }

        if (result.size() > 0 && ((Object[])result.get(0)).length == 1)
        {
            List r = result;
            result = new ArrayList();
            for (int i = 0; i < r.size(); i++)
            {
                result.add(((Object[]) r.get(i))[0]);
            }
        }
        return result;
    }

    private Object[] result(Object obj)
    {
        Symbol symbol = symtbl.getSymbol(candidateAlias);
        symbol.setValue(obj);
        Object[] r = new Object[result.length];
        for (int i=0; i<result.length; i++)
        {
            r[i] = result[i].evaluate(evaluator);
        }
    
        return r;
    }

    private Object[] result(List set)
    {
        // Store the results set so we can aggregate
        Symbol setSymbol = null;
        if (symtbl.hasSymbol(Symbol.RESULTS_SET))
        {
            setSymbol = symtbl.getSymbol(Symbol.RESULTS_SET);
        }
        else
        {
            setSymbol = new PropertySymbol(Symbol.RESULTS_SET);
            symtbl.addSymbol(setSymbol);
        }
        setSymbol.setValue(set);

        // Retrieve the candidate symbol for later processing
        Symbol candidateSymbol = null;
        if (symtbl.hasSymbol(candidateAlias))
        {
            candidateSymbol = symtbl.getSymbol(candidateAlias);
        }
        else
        {
            candidateSymbol = new PropertySymbol(candidateAlias);
            symtbl.addSymbol(candidateSymbol);
        }

        Object[] r = new Object[result.length];
        for (int j=0; j<result.length; j++)
        {
            Object element = set.get(0); // Use first element only. Should be same in other "group-by" components
            candidateSymbol.setValue(element);
            r[j] = result[j].evaluate(evaluator);
        }

        return r;
    }

    /**
     * Constructs ResultClassMapper and calls its map function
     * @param resultSet The resultSet containing the instances handled by setResult
     * @return The resultSet containing instances of the Class defined by setResultClass
     */
    abstract Collection mapResultClass(Collection resultSet);
}