/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License. 
 

Contributors:
2003 Erik Bengtson - removed unused import
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - moved statements to AbstractSetStore
    ...
**********************************************************************/
package org.datanucleus.store.mapped.scostore;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.StateManager;
import org.datanucleus.exceptions.NucleusUserException;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.DiscriminatorStrategy;
import org.datanucleus.store.mapped.DatastoreClass;
import org.datanucleus.store.mapped.DatastoreContainerObject;
import org.datanucleus.store.mapped.DatastoreIdentifier;
import org.datanucleus.store.mapped.IdentifierType;
import org.datanucleus.store.mapped.expression.LogicSetExpression;
import org.datanucleus.store.mapped.expression.QueryExpression;
import org.datanucleus.store.mapped.expression.ScalarExpression;
import org.datanucleus.store.mapped.mapping.JavaTypeMapping;
import org.datanucleus.store.scostore.MapStore;
import org.datanucleus.util.ClassUtils;

import java.util.Collection;

/**
 * Representation of the backing store for a MapValue Set.
 * This is used where the user calls Map.values() and then wants to perform
 * some operation on the resulting Collection.
 * TODO Cater for embedded values.
 */
public abstract class MapValueSetStore extends AbstractSetStore
{
    protected final MapStore mapStore;
    protected final JavaTypeMapping keyMapping;

    private DatastoreIdentifier valueIdentifier;

    /**
     * Constructor where a join table is used to store the map relation.
     * @param mapTable Table used by the map
     * @param mapStore Store used by the map
     * @param clr The ClassLoaderResolver
     * @param ownerMapping the owner mapping
     * @param keyMapping the key mapping
     * @param valueMapping the value mapping
     * @param specialization the specialization
     */
    public MapValueSetStore(DatastoreContainerObject mapTable, MapStore mapStore, ClassLoaderResolver clr, JavaTypeMapping ownerMapping,
                            JavaTypeMapping keyMapping, JavaTypeMapping valueMapping, AbstractSetStoreSpecialization specialization)
    {
        super(mapTable.getStoreManager(), clr, specialization);

        this.containerTable = mapTable;
        this.mapStore = mapStore;

        setName = "values";
        this.ownerMapping = ownerMapping;
        this.keyMapping = keyMapping;
        this.elementMapping = valueMapping;

        elementType = elementMapping.getType();
        elementsAreEmbedded = isEmbeddedMapping(elementMapping);
        elementsAreSerialised = isEmbeddedMapping(elementMapping);

        initialize(clr);

        postInit();
    }

    /**
     * Constructor where either the key is stored in the value table or the value is stored in the key table.
     * @param valueTable Table for the values
     * @param ownerMapping Mapping to the owner
     * @param valueMapping The mapping for this table to the value table/column
     * @param mapStore Store used by the map.
     * @param clr The ClassLoaderResolver
     * @param fmd FieldMetaData for the owning "map" field
     * @param specialization the specialization
     */
    public MapValueSetStore(DatastoreClass valueTable, JavaTypeMapping ownerMapping, JavaTypeMapping valueMapping, MapStore mapStore, 
            ClassLoaderResolver clr, AbstractMemberMetaData fmd, AbstractSetStoreSpecialization specialization)
    {
        super(valueTable.getStoreManager(), clr, specialization);

        this.containerTable = valueTable;
        setName = "values";
        this.ownerMemberMetaData = fmd;

        this.ownerMapping = ownerMapping;
        this.keyMapping = null;
        this.elementMapping = valueMapping;

        this.mapStore = mapStore;

        initialize(clr);

        postInit();
    }

    /**
     * Initialise Method.
     **/
    private void initialize(ClassLoaderResolver clr)
    {
        elementType = elementMapping.getType();
        elementsAreEmbedded = isEmbeddedMapping(elementMapping);
        elementsAreSerialised = isEmbeddedMapping(elementMapping);

        Class element_class = clr.classForName(elementType);
        if (ClassUtils.isReferenceType(element_class))
        {
            emd = storeMgr.getOMFContext().getMetaDataManager().getMetaDataForImplementationOfReference(element_class,null,clr);
        }
        else
        {
            emd = storeMgr.getOMFContext().getMetaDataManager().getMetaDataForClass(element_class, clr);
        }
        if (emd != null)
        {
            elementType = emd.getFullClassName();
            elementInfo = getElementInformationForClass();
        }

        valueIdentifier = storeMgr.getIdentifierFactory().newIdentifier(IdentifierType.TABLE, "VALUE");
    }

    /**
     * Accessor for the statement for the iterator.
     * @param ownerSM the owner StateManager
     * @return The iterator Query Statement.
     **/    
    protected QueryExpression getIteratorStatement(StateManager ownerSM)
    {
        final ClassLoaderResolver clr = ownerSM.getObjectManager().getClassLoaderResolver();
        final Class elementCls = clr.classForName(elementType);
        QueryExpression stmt = null;
        if (emd != null &&
            emd.getDiscriminatorStrategy() != null &&
            emd.getDiscriminatorStrategy() != DiscriminatorStrategy.NONE &&
            containerTable instanceof DatastoreClass)
        {
            // Inverse Map where Value is PC and uses discriminator
            if (ClassUtils.isReferenceType(clr.classForName(elementType)))
            {
                // Take the metadata for the first implementation of the reference type
                String[] clsNames = storeMgr.getOMFContext().getMetaDataManager().getClassesImplementingInterface(elementType, clr);
                Class[] cls = new Class[clsNames.length];
                for (int j=0; j<clsNames.length; j++)
                {
                    cls[j] = clr.classForName(clsNames[j]);
                }
                stmt = getSpecialization().newDiscriminatorIteratorStatement(clr,
                    cls, true, true).getQueryStatement(null);
            }
            else
            {
                stmt = getSpecialization().newDiscriminatorIteratorStatement(clr, new Class[] {elementCls},
                    true, true).getQueryStatement(null);
            }

            iterateUsingDiscriminator = true;

            DatastoreClass sourceTable = storeMgr.getDatastoreClass(elementType, clr);
            if (containerTable != sourceTable)
            {
                // Map using join so add inner join to the container table
                ScalarExpression sourceExpr = 
                    sourceTable.getIDMapping().newScalarExpression(stmt, stmt.getMainTableExpression());
                LogicSetExpression teTargetElement = stmt.newTableExpression(containerTable, valueIdentifier);
                ScalarExpression targetExpr = elementMapping.newScalarExpression(stmt, teTargetElement);
                stmt.innerJoin(sourceExpr, targetExpr, teTargetElement, true);
            }
        }
        else
        {
            stmt = getSpecialization().newUnionIteratorStatement(clr, elementCls, true,
                elementCls, elementMapping, containerTable, false, null, 
                true, false).getQueryStatement(null);
        }

        // Apply condition on owner field to filter by owner
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getMainTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, ownerSM.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);

        return stmt;
    }

    /**
     * Method to add a value to the Map. Not supported.
     * @param sm State Manager for the owner
     * @param element The value to add
     * @return Whether it was added correctly.
     */
    public boolean add(StateManager sm, Object element, int size)
    {
        throw new UnsupportedOperationException("Cannot add to a map through its value set");
    }

    /**
     * Method to add entries to the Map. Not supported.
     * @param sm State Manager for the owner
     * @param elements The values to add
     * @return Whether it was added correctly.
     */
    public boolean addAll(StateManager sm, Collection elements, int size)
    {
        throw new UnsupportedOperationException("Cannot add to a map through its value set");
    }

    /**
     * Method to remove a value from the Map.
     * @param sm State Manager for the owner
     * @param element The value to remove
     * @return Whether it was removed correctly.
     */
    public boolean remove(StateManager sm, Object element, int size, boolean allowDependentField)
    {
        if (!validateElementForReading(sm, element))
        {
            return false;
        }

        return remove(sm, element);
    }

    /**
     * Method to remove elements from the map via the value collection.
     * @param sm StateManager for the owner
     */
    public boolean removeAll(StateManager sm, Collection elements, int size)
    {
        throw new NucleusUserException("Cannot remove elements from a map through its value set");
    }

    /**
     * Method to clear the map.
     * @param sm StateManager for the owner
     */
    public void clear(StateManager sm)
    {
        if (canClear()) {
            throw new NucleusUserException("Cannot clear an inverse map through its value set");
        }

        super.clear(sm);
    }
 
    // --------------------------- JDOQL Query Methods -------------------------------

    public ScalarExpression joinElementsTo(QueryExpression stmt,
                       QueryExpression qs,
                	   JavaTypeMapping ownerMapping,
 		   			   LogicSetExpression te,
 		   			   DatastoreIdentifier setRangeVar,
                       Class filteredElementType,
                       ScalarExpression elmExpr,
                       DatastoreIdentifier elementRangeVar, boolean existsQuery)
    {
        throw new NucleusUserException("Cannot query sets obtained by Map.values()");
    }

    private AbstractSetStoreSpecialization getSpecialization()
    {
        return (AbstractSetStoreSpecialization) specialization;
    }

    protected abstract void postInit();
    protected abstract boolean canClear();
    protected abstract boolean remove(StateManager sm, Object element);

}
